#include <stdio.h>              // printf()
#include <stdlib.h>             // exit()
#include "phys.h"



typedef Bit32 (*reader_t)( void *pointer );



// this set of 3 functions takes care of touching
// only the address and width that is wanted by
// the user
//
// no masking is necessary because C promotion
// will zero out the high bits if necessary
//
Bit32 read32( void *pointer )
{
    return * (Bit32 *) pointer;
}



Bit32 read16( void *pointer )
{
    return * (Bit16 *) pointer;
}



Bit32 read8( void *pointer )
{
    return * (Bit8 *) pointer;
}



int main( int argc, char *argv[] )
{
    Bit32 handle, start, line;
    Bit32 length, length_remaining;
    Bit32 value;
    int width;
    const char *program_name = argv[0];
    Bit8 *pointer, *address;
    reader_t reader;

    // parse command-line arguments
    //
    if( argc != 4 )
    {
        fprintf( stderr, "usage: %s phys-address bit-width length\n", program_name );
        return 1;
    }
    start  = strtoul( argv[1], NULL, 0 );
    length = strtoul( argv[3], NULL, 0 );
    width  = atoi( argv[2] );

    switch( width )
    {
        default:
            fprintf( stderr, "Sorry, you can only use widths of 8, 16, or 32 bits.\n" );
            return 3;

        case 8:
            reader = read8;
            break;

        case 16:
            reader = read16;
            break;

        case 32:
            reader = read32;
            break;
    }

    // map one or more pages of memory starting at the
    // physical address that the user specified
    //
    pointer = (Bit8 *) map_physical( start, length, &handle );
    if( pointer == NULL )
    {
        printf( "Sorry, cannot map %d bytes at physical address 0x%.8X to a flat pointer\n", length, address );
        return 2;
    }

    printf( "dumping %d-bit memory values starting at address 0x%.8X, length 0x%.8X\n", width, start, length );

    for( address=pointer,length_remaining=length; length_remaining>15; address+=16,length_remaining-=16 )
    {
        printf( "%.8X  ", (address-pointer)+start );

        for( line=0; line<16; line+=width/8 )
        {
            value = reader( address + line );
            printf( " %.*X", width/4, value );
        }

        printf( "\n" );
    }

    return 0;
}
