/* 1730, Tue 18 Nov 97

  1555, Thu  1 Oct 98  OC3 test version of new (OCx) meter
  1330, Wed 30 Apr 97  Nevil at ISMA-2, SDSC, San Diego
  2245, Thu  5 Dec 96  First OC3MON version, MCI, Reston

NeTraMet outer block, moved into OC3MON environment.

OC3MON collects tracerecs (one for the first cell of each
packet), and passes them to routines in fs2flows.c.
This module provides the same interface calls as fs2flows.

For OC3MON, we use only a single Ethernet interface, which
handles IP traffic to/from the meter.  It doesn't do any
monitoring of the Ethernet!

*/

/* $Log: meter_oc.c,v $
 *
 * Revision 1.3  1998/05/20 04:02:28  rtfm
 * Declare startup = 1; otherwise meter doesn't start properly
 */


#define  LINK_HAS_LLC_SNAP    TRUE 
#define  ASSUME_ETHERTYPE_IS_IP  0  /* Get type from SNAP header */

#include <sys/types.h>
#include <dos.h>
#include <conio.h>
#include <alloc.h>
#include <ctype.h>


#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <float.h>              /* DBL_MIN DBL_MAX */

#if __unix__
#   include <sys/time.h>
#   include <sys/resource.h>
#   include <netinet/in.h>
#else
#   include "..\wattcp\tcp.h"
#endif

#if __unix__
#   include "types.h"
#else
#   include "..\types\types.h"
#   include "..\intel\timestmp.h"
#   include "..\intel\byteswap.h"
#endif

#include "..\integrat\trace.h"  /* must be included before util.h */
#include "..\integrat\buffer.h"             /* class BufferBlock */
#include "..\integrat\fs2flows.h"  /* make sure we comply with our own prototypes */

#if 0
#include "queue.h"
#include "util.h"
#include "dostats.h"
#endif

#define EXTSNMP  /* Declare au_snmp_port */
#include "ausnmp.h"

/*  #include "..\wattcp\pktdrv.h"  !! */

#include "..\snmplib\asn1.h"
#include "..\snmplib\snmp.h"
#include "..\snmplib\snmpimpl.h"
#include "..\snmplib\snmpapi.h"

#define PKTSNAP
#include "pktsnap.h"
#include "flowhash.h"
#include "met_vars.h"

extern char version_descr[];  /* In met_vars.c */
extern char *communities[];  /* In snmp/snmpagnt.c */

extern int errno;
int snmp_dump_packet = 0;

udp_Socket udpsock;
static udp_Socket *snmp_sp;
int status;  /* For socket */
unsigned long gethostid(void);  /* In src\pcbsd.c */
void init_snmp(void);  /* In snmp\snmpapi.c */

unsigned char half_scale = 0;

void zero_pkt_stats(void);  /* meter_oc function prototypes */
void receive(void);
void meter_snmp_read(int length);


u_int active_feed = 1;
PentiumClock active_alarm;


#define MXINTERFACES  4  /* Max nbr of interfaces to meter */

struct interface_info {
   u_char nbr;  /* Interface number (for Surce/DestInterface attributes) */
   char name[20];  /* Interface name, e.g. nf0 */
   u_int SampleRate,  /* 1 = count every packet on this interface */
      sample_count;
   u_int ipackets;  /* Total packets since last stats reset */
   u_int idrops;  /* Dropped packets since last stats reset */
   };
struct interface_info pci[MXINTERFACES];

int n_interfaces = 2;  /* Nbr of active interfaces */
char interface_name[10] = "oc3a";  /* Device name for metered interface */


// none of this code gets compiled in
// if it's isn't going to be called
//
#if DO_FOREGROUND_STATS

void usage(char *prog)
{
   fprintf(stderr, "Usage: %s -u mxrules [-f mxflows]\n", prog);
   fprintf(stderr, "   -w wrcom [-r rdcom]\n");
   }


void bump_noflowpkts(int if_nbr)
{  }

void startup_screen()
{
   char cbuf[60];
   sclear();
   scpos(0,0);  printf(version_descr);
   scpos(0,1);  printf("Running on %s",inet_ntoa(cbuf,gethostid()));
   }
   
// Called once (on startup).  Passes command line in
//    (OC3MON doean't have any command-line parameters)
//
int flows_init(
   int argc,
   char *argv[],
   PentiumClock *keystroke_interval,
   PentiumClock *query_interval,
   PentiumClock *active_interval
   )
{
   int arg, c, sample_rate;
   char cbuf[20], *ap;
   struct en_info *enp;

   if (argc < 2) {
      usage(argv[0]);
      return 1;
      }

   scinit();  sclear();  /* Set screen size parameters */
   display_enabled = kb_enabled = 1;  /* Enabled by default */
   chart_interface = 0;
   mxflows = DFMXFLOWS;  mxrules = DFMXRULES;
   /* Default (CMU) communities are:  0 = "public", 1 = "proxy",
     2 = "private", 3 = "regional", 4 = "core"
      We only allow "public" and "private" by default */
   communities[1] = communities[3] = communities[4] = "";

   n_interfaces = 2;  /* For OC3MON */
   memset(pci, 0, sizeof(pci));
   pci[0].nbr = 1;  pci[1].nbr = 2;
   pci[0].SampleRate = pci[1].SampleRate = 1;
   pci[0].sample_count = pci[1].sample_count = 1;

   for (arg = 1; arg < argc; ++arg) {
      if (argv[arg][0] == '-') {
         ap = argv[arg]+2;
         switch (argv[arg][1]) {
         case 'd':
            snmp_dump_packet++;
            break;
         case 'f':
            if (*ap == '\0') ap = argv[++arg];
            mxflows = atoi(ap);
            break;
         case 'k':
            kb_enabled = 0;  /* -k to disable keyboard */
            break;
         case 'm':
            if (*ap == '\0') ap = argv[++arg];
            au_snmp_port = atoi(ap);
            break;
         case 'n':
            if (*ap == '\0') ap = argv[++arg];
            sample_rate = atoi(ap);
            if (sample_rate < 1) sample_rate = 1;
            pci[0].SampleRate = pci[1].SampleRate = sample_rate;
            break;
         case 'r':
            if (*ap == '\0') ap = argv[++arg];
            communities[0] = ap;  /* -r to set read community */
            break;
         case 's':
            display_enabled = 0;  /* -s to disable screen */
            break;
	 case 'u':
	    if (*ap == '\0') ap = argv[++arg];
	    mxrules = atoi(ap);
	    break;
         case 'w':
            if (*ap == '\0') ap = argv[++arg];
            communities[2] = ap;  /* -w to set write community */
            break;
         default:
            printf("Invalid option: -%c\n", argv[arg][1]);
            return 1;  /* Error */
            }
         }
      }

   snmp_sp = &udpsock;  /* myport        hisadr hisport */
   if (!udp_open( snmp_sp, SNMP_PORT,  0, 0,        0 )) {
      puts("udp_open failed!\n");
      return 1;  /* Error */
      }

   init_snmp();
#if NF_PC_BGP
   InitSubnet();
   bgp_result = read_bgp_file("\\coral\\bin\\bgp.txt" );
   if (bgp_result) {
      printf(">>>> read_bgp_file() returned %d <<<<", bgp_result);
      exit(99);
      }
#endif

   init_monitor(1);

   // Set an "alarm" interval when we will
   // query the statistics
   //
   active_interval->put_seconds( 1.0 );  /* For housekeeping */

   query_interval->put_seconds( 0.02 );  /* For checks on WATTCP */

   keystroke_interval->put_seconds( 1.0 );  /* For checks on keyboard */

   return 0;
   }


// Clears all statistics after being queried remotely
// or during initialization of module
//
// Note that we do not touch the report timestamp
// so that querying will not delay timing out a flow
//
int flows_clear(void)
{
   return 0;
   }

unsigned long nbytes;  /* npackets declared in pktsnap.h */

struct ether_hdr {
   unsigned char dest[6];    /* MAC addresses */
   unsigned char source[6];
   unsigned char type[2];    /* Length or protocol type */
   };

// Called in foreground by main data capture program
//
int flows_per_block(BufferBlock *block)
{
   struct pkt pp;
   unsigned int ethertype, lsap;
   struct ether_hdr ehdr;

   int final_result = 0;
   unsigned count;
   tracerec *record;

   int ifnbr;
   struct interface_info *pcip;
#if DEBUG_COPY_PACKET
   tracerec debug_record;
#endif

#if CHECK_USED_OVERFLOW
   // safety check: used count cannot be greater
   // than COMPL_RING_ENTRIES
   //
   if (block->used > COMPL_RING_ENTRIES) {
      printf("Internal error: block used %d > %d max!\n",
         block->used, COMPL_RING_ENTRIES);
      return 1;
      }
#endif

   memset( &ehdr, 0, sizeof(ehdr) );

   /* use interface number as most significant byte
    * of destination MAC address
    *
    * do this outside the loop since all packets in
    * the block are for the same interface card
    * (direction of data flow)
    *
    * change from zero-based to one-based numbering,
    * so the interface number stands out from the
    * zero padding
    */
   ehdr.dest[0] = ifnbr = block->cardindex + 1;
   pcip = &pci[ifnbr-1];

   /* for all packets headers in the block
    */
   for (count = block->queried; count<block->used; count++) {
      record = (tracerec *)(block->buffers + count*BUFFER_SIZE);

      if (--pcip->sample_count == 0) {  /* Process only 1 pkt per SampleRate */
         pcip->sample_count = pcip->SampleRate;

#if DEBUG_COPY_PACKET
         // copy the record because the debugger, for some
         // idiotic reason does not have the same addressibility
         // that our program has
         //
         debug_record = *record;
#endif

         /* use UNI VPI and VCI as low 3 bytes (4 bytes if we assume
          * NNI) of destination MAC address
          */
         * (Bit32 *) &ehdr.dest[2] = htonl( VPI_VCI( record->atm.whole ) );

#if !ASSUME_ETHERTYPE_IS_IP
         lsap = record->tr_llc.lsap;
         if (lsap == LSAP_SNAP) {
            ethertype = ntohs(record->tr_snap.ethertype);
            if (ethertype == ETHERTYPE_IP)
// old version   pp.p_len = ntohs(record->tr_len);  /* Length from IP header */
               pp.p_len = ntohs(record->tr_ip.ip_len);  /* Length from IP header */
                /* Joel's tr_len define requires first->, which is part
                   of a cellhashblock structure ??   Nevil, 28 Jan 98 */
            else
                pp.p_len = 0;
            }
         else { /* Other type of packet */
            ethertype = 0;
            pp.p_len = 0;
            }
#else
         lsap = 0xAAAA;  ethertype = 0x0800;
#endif

         ++npackets;  nbytes += pp.p_len;

/*       block->cardindex = ATM card nbr !! */
         pp.Low.Interface = pp.High.Interface =  ifnbr;
            /* NeTraMet-style interface number */
         pp.Low.AdjType = pp.High.AdjType = AT_AAL5;  /* ATM over AAL5 */
         record->when.strip_bitfields();
         pp.arrival_time = record->when.get_seconds();

         handle_pkt(&pp, ethertype,lsap,
            (unsigned char *)&ehdr,   /* Ethernet header (for MAC addresses) */
            (unsigned char *)&record->tr_ip,  /* Start of IP header */
            40  /* This is an (OC3MON) dummy for Bytes in buffer */
            );
         pkt_monitor(&pp);
         }

      }   // end for all records in block

   // remember how much of this block we queried, in
   // case it is a partial that we will see again later
   //
   block->queried = count;

   return final_result;
   }


static n_waiting = 0;  /* Chars waiting in snmp socket */

// OC3MON calls this every query_interval seconds, so as to keep wattcp alive
// If it gets a non-zero result, it will call flows_query (below)
// to do whatever socket processing is required.
//
// nonzero return causes main() to ask cards how much of any
// pending block they have filled in, call flows_per_block()
// for those partial blocks, and then call flows_query()
//
int flows_is_query_wanted(void)  /* Interface changed $$$ */
{
   if (!tcp_tick(snmp_sp)) {
      fprintf(stderr,"flows_is_query_wanted(): tcp_tick failed!!\n");
      return 0;
      }
   n_waiting = sock_dataready(snmp_sp);
   return n_waiting != 0;
   }


// Called when program is queried remotely (see above)
//
int flows_query(const char *extra_msg)  /* Interface changed $$$ */
{
   meter_snmp_read(n_waiting);
   sock_close(snmp_sp);
   if (!udp_open(snmp_sp, SNMP_PORT,  0, 0,   0 )) {
      puts("udp_open failed!\n");
      return 1;  /* Error result */
      }
   return 0;  /* Success */
   }




// Called when program is about to exit
//
int flows_cleanup(void)
{
   return 0;
   }



unsigned int startup = 1, p;
int gci;  /* Check this in meter_pc (was char) !!! */
char l_tod_s;  /* Last tod_s displayed on chart */
unsigned int sumx, minx, maxx, samples;  /* Chart-building data */
unsigned long pd,bd;


unsigned int bkgi;  /* Seconds before next run of background process */
#define BKG_INTERVAL  30  /* 30 seconds */


void show_meter_time()
{
   char msg[60];
   sprintf(msg,"%lu seconds since %02d%02d:%02d",
      elapsed_sec, s_tod_h,s_tod_m,s_tod_s);
   display_msg(0,msg);
   }

void zero_pkt_stats()
{
   int j;
   npackets = nbytes = 0L;
   badpackets = nobufpackets = lostpackets  =
      stats_time = t_backlog = spackets = sbytes =
      kilodummypackets = dummypacketrate = 0L;
   mindummyrate = 100000000L;
   pkt_backlog = max_pkt_backlog = max_pkt_rate =
      dummypackets = mdpacketrate = 0;
   clear_pkt_stats = 0;
   }


// Called by main() every active_interval
// OC3MON could use this to send regular data via a
// (possibly tunneled) UDP (possibly multicast) stream
//
// NeTraMet uses it for its 'one-second' processing
//
void flows_send_active(void)
{
   unsigned int x, ax;
   char msg[60];
   struct mgr_rec *mip;
   struct rdr_rec *cip;

   ++elapsed_sec;  set_tod();
   if (startup) {  /* Running now, initialise counters etc. */
      clear_pkt_stats = 1;
      minx = 30000;  sumx = samples = maxx = 0;
      spackets = npackets = 0;  sbytes = nbytes = 0;
      l_tod_s = -1;
      gci = gc_interval;  /* Garbage collect interval */
      bkgi = BKG_INTERVAL;  /* Background interval */
      gc_f = mxflows/100;  /* Flow indexes to check */
      if (gc_f < 4) gc_f = 4;
      startup = 0;

      startup_screen();
      scpos(0,scr_mtrow);
         printf("Room for %d flows, sizeof(flow) = %d",
            mxflows, sizeof(struct flow));
      }
   else {
#ifdef NEW_ATR
      check_rates(uptime_s());
#endif
      if (--gci == 0) {
         garbage_collect(1);  /* Routine incremental collection */
         gci = gc_interval;
         }
      if (--bkgi == 0) {  /* Check % of flows active */
         p = (unsigned long)active_flows()*100/mxflows;
         if (display_enabled) {
            scpos(12,3);  printf("a=%3u",p);
            }
         if (p > FloodMark && FloodMode != TV_TRUE &&
               FloodMark > 0 && FloodMark < 100) {
            FloodMode = TV_TRUE;  /* Stop creating new flow records */
            display_msg(1,"Meter in Flood mode");
            }
         else for (x = 0; x != sizeof(mi)/sizeof(struct mgr_rec); ++x) {
            mip = &mi[x];
            if (mip->mi_Status != RS_ACTIVE
                  || mip->mi_RunningStandby == TV_TRUE)
               continue;
            else if (p > mip->mi_HighWaterMark &&
              mip->mi_HighWaterMark > 0 &&
                  mip->mi_HighWaterMark < 100 &&
                  mip->mi_StandbyRuleSet != 0 &&
                  mip->mi_StandbyRuleSet != mip->mi_CurrentRuleSet &&
                  (mip->mi_StandbyRuleSet != 0 &&
                     ri[mip->mi_StandbyRuleSet-1].ri_Status == RS_ACTIVE)
                  ) {
               open_rule_set(mip->mi_CurrentRuleSet, 0);  /* Close */
               open_rule_set(mip->mi_StandbyRuleSet, 1);  /* Open */
               mip->mi_RunningStandby = TV_TRUE;
               sprintf(msg,"Manager %d running Standby", x+1);
               display_msg(1,msg);
               }
            else if (p > HighWaterMark*2/3)
           more_garbage();
            }
         pd = uptime_cs();
         for (x = 0; x != sizeof(ci)/sizeof(struct rdr_rec); ++x) {
            cip = &ci[x];
            if (cip->ci_Status != RS_ACTIVE || cip->ci_Timeout == 0)
               continue;
            if ((pd - cip->ci_LastTime)/100 > cip->ci_Timeout)
               cip->ci_Status = RS_UNUSED;  /* Timed out */
            }
         bkgi = BKG_INTERVAL;
         }

      pd = (npackets-spackets)*pci[0].SampleRate;
      bd = (nbytes-sbytes)*pci[0].SampleRate;
      if (display_enabled) {
/*         scpos(0,3);
         printf("q=%4d",pkt_backlog);  ?? */
         scpos(23,3);
         printf("%02d%02d:%02d", tod_h,tod_m,tod_s);
         scpos(0,5);
/*         x = util_pc(pd,bd);  /* % utilisation ?? */

         x = bd / 125000;  /* Mbps */
         printf("p=%6lu, b=%10lu, u=%4d", pd,bd, x);
         sumx += x;  ++samples;
         if (x < minx) minx = x;
         if (x > maxx) maxx = x;
         if (tod_s != l_tod_s && tod_s%10 == 0) {
            if (samples) {  /* 30% in col 79 */
               ax = (sumx+(samples>>1))/samples;
               if (half_scale)
                     chart(scr_chrtcol,0, scr_lcol,scr_lrow,
                  minx>>1,ax>>1,maxx>>1);
               else chart(scr_chrtcol,0, scr_lcol,scr_lrow,
                  minx,ax,maxx);
               l_tod_s = tod_s;
               }
            minx = 30000;  sumx = samples = maxx = 0;
            }
         }

      }
   if (clear_pkt_stats) zero_pkt_stats();
   else {
      ++stats_time;
      if (pd > max_pkt_rate) max_pkt_rate = pd;
      t_backlog += (unsigned long)pkt_backlog;
      if (pkt_backlog > max_pkt_backlog)
         max_pkt_backlog = pkt_backlog;
      if (dummypacketrate < mindummyrate) {
         mindummyrate = dummypacketrate;  mdpacketrate = pd;
         }
      }

   pkt_backlog = 0;  dummypacketrate = 0;
/*   spackets = eip->npackets;  sbytes = eip->nbytes;  !! */
   spackets = npackets;  sbytes = nbytes;
   }



unsigned long snmp_peer_addr;

void meter_snmp_read(int length)
{
   int out_length;
   Bit8 packet[1520], outpacket[1550];  /* Leave some space just in case */
   int count;
   struct sockaddr from;
   int fromlen = sizeof(struct sockaddr);
   char snmp_peer_name[30];  /* Name of host which sent the snmp request */
   getpeername((tcp_Socket *)snmp_sp, &from,&fromlen);
   snmp_peer_addr = from.s_ip;

   sock_read(snmp_sp, packet,length);

   if (snmp_dump_packet) {
      inet_ntoa(snmp_peer_name,from.s_ip);
      printf("received %d bytes from %s:\n", length, snmp_peer_name);
      for (count = 0; count < length; count++) {
         printf("%02X ", packet[count]);
         if ((count % 16) == 15) printf("\n");
         else if ((count % 4) == 3) printf(" ");
         }
      printf("\n\n");
      }
   out_length = 1500;
   if (snmp_agent_parse(packet, length, outpacket, &out_length, from.s_ip)) {
      if (snmp_dump_packet) {
         printf("sent %d bytes to %s:\n", out_length, snmp_peer_name);
         for (count = 0; count < out_length; count++) {
            printf("%02X ", outpacket[count]);
            if ((count % 16) == 15) printf("\n");
            }
         printf("\n\n");
         }
      sock_write(snmp_sp, (char *)outpacket, out_length);
      }
   }

void flows_keystroke(int keystroke)
{
   w_clear(0,scr_lrow, scr_chrtcol-1,scr_lrow);  /* Status area */
   scpos(0,scr_lrow);
   switch (keystroke) {
   case 'h':
      half_scale = !half_scale;
      printf("Chart scale 0..%d Mbps",
      half_scale ? 60 : 30);
      break;
   case 'm':
      printf("%lu bytes free", farcoreleft());  /* No near heap */
      break;
   case 'v':
      printf(version_descr);
      break;
   case '=':
      startup_screen();  /* Clear the clutter off the screen */
      break;
   default:
      handle_kb(keystroke);
      break;
      }
   }


#endif  // DO_FOREGROUND_STATS

