/* 1804, Thu 20 Aug 98

   PKTSNAP.H:  Declarations for AU packet monitoring
          extensions to Waterloo TCP/IP

   Copyright (C) 1992-1998 by Nevil Brownlee,
   ITSS Technology Ddevelopment,  The University of Auckland */

/* $Log: pktsnap.h,v $
 *
 * Revision 1.1  1998/05/07 04:13:44  rtfm
 * Initial revision
 */

#if TIME_WITH_SYS_TIME
#include <time.h>
#include <sys/time.h>
#else
#if HAVE_SYS_TIME_H
#include <sys/time.h>
else
#include <time.h>
#endif
#endif

#ifdef PKTSNAP
#define EXTERN
#else
#define EXTERN extern
#endif

#if defined(NETFLOW)
#  if (NETFLOW & 0x01) != 0
#     define NF_ASN_ATT      1
#  endif
#  if (NETFLOW & 0x02) != 0
#     define NF_OTHER_ATT    1
#  endif
#  if (NETFLOW & 0x04) != 0
#     define NF_CISCO_DATA   1
#  endif
#  if (NETFLOW & 0x08) != 0
#     define NF_PC_BGP       1
#  endif
#endif 

#define noFULL_IPX  /* Net+Host => 10-byte IPX peer addresses */
#define noCLNS      /* CLNS peer addresses are 20-byte NSAPs */

#if defined(FULL_IPX) || defined(CLNS)
#define BIG_PEER_ADDR  /* 20-byte peer addresses */
#endif

/* Determining snapshot length:  (Nevil, 9 Apr 98)

   Worst case IP packet:       Minimal IP packet:
        14  Ethernet                14  Ethernet
         8  802.2 + SNAP             8  802.2 + SNAP
        60  IP header               20  IP header
        20  TCP header              20  TCP header
      ----                       -----
       102  bytes                   62  bytes

   Leaving out Ethernet and 802.2 headers (22 bytes),
   the peer header lengths are:
      IP4:   80 bytes
      CLNS:  53 bytes  (12 header + 2x21 NSAPs)  */

#define MAC_ADDR_LEN     6  /* Address lengths */
#define SNAP_MAC  2*MAC_ADDR_LEN

#if defined(SOLARIS23)
#define SNAPSIZE   54L  /* Enough for all except CLNS (which needs 70) */
/* SNAP size currently set to 54 to sidestep a bug in Solaris 2.3, 
   which corrupts packets smaller than the SNAP size.  This bug is 
   fixed in Solaris 2.4; if you are using Solaris 2.3 you should
   set SNAPSIZE to MXPKTHDRLEN.   Nevil, 12 Jul 94 */

#elif defined(DOS)
#define SNAPSIZE  54  /* Packet Driver source: 8390.asm */

#elif defined(OCX)
#define SNAPSIZE  SNAP_MAC + 48  /* Kludged MACs + first AAL5 cell */

#else
#define SNAPSIZE 102  /* Worst case IP, as above */
#endif

EXTERN int pktsize;   /* Snapshot - header */

#define IP4_ADDR_LEN     4
#define IP6_ADDR_LEN    16
#define DN_ADDR_LEN      3
#define ET_ADDR_LEN      3
#ifdef FULL_IPX
# define IPX_ADDR_LEN   10
#else
# define IPX_ADDR_LEN    4
#endif
#define NSAP_ADDR_LEN   20

#define TRANS_ADDR_LEN   2

#ifdef BIG_PEER_ADDR
# define PEER_ADDR_LEN  20  /* Default */
#else
# define PEER_ADDR_LEN   4  /* Default */
#endif

#ifdef BIG_PEER_ADDR
# define RULE_ADDR_LEN  20  /* Max of above */
#else
# define RULE_ADDR_LEN   6  /* Max of above */
#endif

struct pkt_key {
   unsigned char AdjAddress[MAC_ADDR_LEN];
   unsigned char PeerAddress[PEER_ADDR_LEN];
   unsigned char TransAddress[TRANS_ADDR_LEN];
   unsigned char Interface, AdjType;
#if NF_ASN_ATT
   unsigned char nf_ASN[TRANS_ADDR_LEN];
#endif
#if NF_OTHER_ATT
   unsigned char nf_mask;
#endif
   };

struct pkt {
   unsigned int next;  /* For linking into queues */

   unsigned char PeerAddrType, TransAddrType;
   struct pkt_key Low, High;

#if NF_CISCO_DATA
   unsigned char ntm_interface;  /* Index in  -i  list */
   unsigned long dPkts, dOctets;
   unsigned long dFirst, dLast;
#else
#define ntm_interface  Low.Interface
   unsigned int p_len;
# if defined(OCX)
   double arrival_time;
# elif defined(DOS)
   counter64 arrival_time;
# else  /* Unix */
   struct timeval arrival_time;
# endif
#endif
   };


#if defined(DOS)

/* Declared in wattcp\src\pcpkt.c .. */

extern unsigned char eth_addr[6];
#endif

/* Declared in wattcp\acct\meter_pc.c .. */

EXTERN unsigned int  /* Statistics variables */
   max_pkt_rate, pkt_backlog,max_pkt_backlog,
   mdpacketrate;
EXTERN int chart_interface;
EXTERN unsigned long
   stats_time, t_backlog,
   spackets,sbytes, zpackets,
   badpackets,nobufpackets;
EXTERN int
   clear_pkt_stats;  /* 1 to clear stats variables */

/* 12 bytes interframe space + 8 bytes preamble + 4 byte FCS */
#define util_pc(p,b)  (((24l*p + b)/1250l + 5l)/10l)  /* % utilisation */

#ifdef DOS  /* Variables for meter 'idle %' statistics */
EXTERN unsigned int dummypackets;
EXTERN unsigned long
   kilodummypackets, dummypacketrate, mindummyrate;
#else
EXTERN clock_t
   realtime, realtime_org, srealtime,
   proctime, proctime_org, sproctime,
   min_idle1000;
#endif

EXTERN int
   scr_lrow,      /* Y co-ord: bottom line of screen */
   scr_mtrow,     /* Y co-ord: top line of status message area */
   scr_lcol,      /* X co-ord: rightmost col of screen */
   scr_chrtcol;   /* X co-ord: leftmost col of chart */

#ifndef DOS
#define w_roll(x1,y1, x2,y2, up)  /* PC screen functions */
#define w_clear(x1,y1, x2,y2)

void start_uptime_clock(void);
Bit32 uptime_s(void);
Bit32 uptime_cs(void);

double centiseconds(struct timeval nt);
double microseconds(struct timeval nt);
#else

#ifdef __cplusplus
extern "C" {  // chart.cpp functions
#endif
   int kbstatus(void);
   unsigned char kbcode(void);
   void bios_scroll(int way, int tx,int ty, int bx,int by, int nrows, int attr);
   void scinit(void);
   void scpos(int x, int y);
   void set_tod(void);
   void sclear(void);
   void swrite(int c);
   void w_clear(int tx,int ty, int bx,int by);
   void w_roll(int tx,int ty, int bx,int by, int n);
   void chart(int tx,int ty, int bx,int by, int x1,int x2,int x3);

   void start_uptime_clock(void);
   Bit32 uptime_s(void);
   Bit32 uptime_cs(void);

#ifdef OCX
   double uptime(void);
   double centiseconds(double s);
   double microseconds(double s);
#else
   counter64 uptime(void);
   double centiseconds(counter64 nt);
   double microseconds(counter64 nt);
#endif
#ifdef __cplusplus
   };
#endif
#endif

/* Monitor variables .. */

EXTERN unsigned long npackets;
EXTERN long lostpackets;

#define MAXPKTLEN  1526

#if defined(DOS)
EXTERN struct pkt far *pkts;
EXTERN int
   maxpkt,       /* Size of meter input buffer */
   free_n_pkt,   /* Available pkt stack */
      free_head,free_tail,
   proc_n_pkt,   /* Queue of pkts to be processed */
      proc_head,proc_tail;
#endif

EXTERN long l_tod;  /* Time of day */
EXTERN int tod_h,tod_m,tod_s,
   s_tod_h,s_tod_m,s_tod_s;  /* Time when counters last saved */

EXTERN long elapsed_sec;  /* Seconds since counters last saved */

#ifndef NULL
#define NULL  0
#endif
