#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "..\intel\timestmp.h"
#include "..\wattcp\wattcp.h"
#include "..\wattcp\elib.h"
#include "..\types\types.h"
#include "..\dpmi32\phys.h"

extern void (*_dbugxmit)(
    sock_type *s, in_Header *ih, tcp_Header *ph, char *line);
extern void (*_dbugrecv)(
   sock_type *s, in_Header *ih, tcp_Header *ph, char *line);

static trace_level;
static PentiumClock startpc;

void p_upticks(void)
{
   PentiumClock now;
   double s;  Bit32 t;
   now.set();  now -= startpc;
   s = now.get_seconds();  t = now.get_ticks();
   printf("%.3f [%4u] ", s, t%10000);
   }

#define TCP_TX  1
#define TCP_RX  2

void p_tcp_pkt(int type,
   tcp_Socket *s, in_Header *ip, tcp_Header *tp, char *line)
{
   int flags = ntohs(tp->flags);
   int len, d_offset;
   Bit32 seq;
   char *msg[] = {"TCP_TX", "TCP_RX"};

   p_upticks();
   printf("%s:   ", msg[type-1]);
   if (flags & tcp_FlagSYN) printf("S");
   if (flags & tcp_FlagACK) printf("A");
   if (flags & tcp_FlagFIN) printf("F");
   if (flags & tcp_FlagRST) printf("R");

   len = in_GetHdrlenBytes(ip);  // Bytes in IP header
   len = ntohs(ip->length) - len;  // Bytes in TCP segment
   d_offset = tcp_GetDataOffset(tp) << 2;  // Bytes in TCP header
   len -= d_offset;  // Bytes in TCP (options and) data

   if (trace_level > 1) printf("   %s\n", line);
   else {
      printf("  len=%u, seq=%lu,   ack=%lu\n",
         len, ntohl(tp->seqnum), ntohl(tp->acknum));
      return;
      }

   seq = ntohl(tp->seqnum);
   printf("   p_len=%u, p_seq=%lu, e_ack=%lu,  p_ack=%lu, p_win=%u\n",
      len,seq,seq+len, ntohl(tp->acknum),ntohs(tp->window));

   printf("   s_seq=%lu, s_unack=%u, s_sdatalen=%u,  s_rdatalen=%u, s_ack=%lu\n",
      s->seqnum, s->unacked, s->sdatalen,
      s->rdatalen, s->acknum);
   }

static void dbugxmit(sock_type *st, in_Header *ih, tcp_Header *ph, char *line)
{
   if (st == NULL) return;
   if (st->tcp.ip_type != IPPROTO_TCP) return;
   p_tcp_pkt(TCP_TX, &st->tcp, ih, ph, line);
   }

static void dbugrecv(sock_type *st, in_Header *ih, tcp_Header *ph, char *line)
{
   if (st == NULL) return;
   if (st->tcp.ip_type != IPPROTO_TCP) return;
   p_tcp_pkt(TCP_RX, &st->tcp, ih, ph, line);
   }

void init_trace(int level)
{
   if (level == 0) return;
   startpc.set();
   _dbugxmit = dbugxmit;
   _dbugrecv = dbugrecv;
   trace_level = level;
   }
